/*	================================================================================
	Project		: TLS_test
	File		: TLS_test.c
	Copyright	: 1997 - Laurent POINTAL
	--------------------------------------------------------------------------------
	Just a code to test TLS.
	NOTE: TLS functions calls are in assert() calls... so you must compile in
		debug mode, or change the code to keep an error checking.
	================================================================================	*/

#include <KernelKit.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <devtools/TLS.h>

//	Change counts to test.
#define NUMBER_OF_CLIENTS			10
#define NUMBER_OF_INDEXS			100
#define NUMBER_OF_CHECKS			5

//	Indexs allocated.
int indexThreadStr ;
int indexs[NUMBER_OF_INDEXS] ;

//	Threads created.
sem_id main_locker = 0 ;
int32  client_term_count = 0 ;

//	Test functions...
int main (int argc, char** argv, char** envp) ;
int32 task (void* p) ;
void doTaskJob (int* datas) ;
void entryProc (thread_id id, int param, void* threadEnv) ;
void exitProc (thread_id id, int param, void* threadEnv) ;


/*	===== main =====================================================================
	Main program.
	================================================================================	*/
int main (int argc, char** argv, char** envp)
	{
	int i ;
	char name[40] ;
	thread_id id ;
	status_t status ;
	
	printf ("\n\nTest of Thread Local Storage sources.\n") ;
	printf ("(C)1997 - Laurent Pointal.\n") ;
	printf ("Parameters: number of clients       = %d\n", NUMBER_OF_CLIENTS) ;
	printf ("            number of indexs        = %d\n", NUMBER_OF_INDEXS) ;
	printf ("            number of values check  = %d\n", NUMBER_OF_CHECKS) ;
	printf ("(these parameters are #defined and can be modified in sources)\n") ;
	printf ("Will allocate indexs, and create clients. Then wait for end of all clients.\n") ;
	printf ("Each client will allocate random values for each index and store it using\n") ;
	printf ("TLS and using its own stack-private array. Then it will call a check \n") ;
	printf ("function which get the client private values and compare it to the\n") ;
	printf ("TLS stored ones. Between two tests, clients will sleep. Clients has\n") ;
	printf ("different priorities.\n") ;
	printf ("Starting and ending of clients threads, and entry/exit TLS procs are traced.\n\n") ;
	printf ("NOTE: This is a hard resource contention test (all clients are only doing TLS\n") ;
	printf ("      data access). You can see this with TManager.\n") ;
	printf ("Press return to start.\n") ;
	getchar () ;

	main_locker = create_sem (0, "main locker") ;
	
	printf ("Initializing TLS.\n") ;
	assert (B_NO_ERROR == tls_init (TLS_VERSION)) ;
	
	printf ("Allocating indexs.\n") ;	
	assert (B_NO_ERROR == tls_alloc (&indexThreadStr)) ;
	for (i=0 ; i<NUMBER_OF_INDEXS ; i++)
		assert (B_NO_ERROR == tls_alloc (&indexs[i])) ;
	assert (B_NO_ERROR == tls_install_entry_proc (entryProc, indexThreadStr)) ;
	assert (B_NO_ERROR == tls_install_exit_proc (exitProc, indexThreadStr)) ;

	printf ("Creating clients.\n") ;
	//	Create client threads, with different priority levels.
	for (i=0 ; i<NUMBER_OF_CLIENTS ; i++)
		{
		sprintf (name, "Task(%d)", i+1) ;
		id = spawn_thread (task, name, B_NORMAL_PRIORITY+rand()/2048, (void*)(i+1)) ;
		resume_thread (id) ;
		}
	
	printf ("Running...\n") ;
	//	Wait for end of all child threads.
	while (client_term_count < NUMBER_OF_CLIENTS)
		acquire_sem (main_locker) ;
	delete_sem (main_locker) ;
	
	//	Now, can terminate tls services.
	printf ("Terminating TLS.\n") ;
	assert (B_NO_ERROR == tls_term ()) ;
	
	printf ("Terminating test program.\n") ;
		
	return 0 ;
	}


/*	===== task =====================================================================
	The thread entry point.
	It allocate random datas and store them using TLS and using a local array on
	the stack.
	Then it periodically call a function which retrieve the datas stored by
	TLS and compare them with the datas stored locally.
	================================================================================	*/
int32 task (void* p)
	{
	int i ;
	int myIndex = ((int) p)+1 ;
	long long tempo ;
	int id = find_thread (NULL) ;
	int datas[NUMBER_OF_INDEXS] ;

	//	To have different snooze for each thread.
	tempo = 10LL * rand() ;

	printf ("Thread(%d) started. Snooze=%f sec\n", id, (double)tempo/1000000.0) ;
	
	//	Half threads will call explicitly entering/exiting procs.
	if (id%2==1) assert (B_NO_ERROR == tls_entering_procs ()) ;

	//	Creating and storing datas...
	for (i=0 ; i<NUMBER_OF_INDEXS ; i++)
		{
		datas[i] = rand () ;	// 0 to 32767
		assert (B_NO_ERROR == tls_set (indexs[i], datas[i])) ;
		}

	//	Testing if datas are well stored.
	for (i=0 ; i<NUMBER_OF_CHECKS ; i++)
		{
		doTaskJob (datas) ;
		snooze (tempo) ;
		}
	printf ("Thread(%d) terminated.\n", id) ;

	//	Half threads will call explicitly entering/exiting procs.
	if (id%2==1) assert (B_NO_ERROR == tls_exiting_procs ()) ;
	
	//	Signal main thread that another one child has finished working.
	atomic_add (&client_term_count, 1); 
	release_sem (main_locker) ;
	return 0 ;
	}


/*	===== doTaskJob ================================================================
	The function doing its job.
	Note that it dont remain the values, it get them from caller, and check if
	they are the same as the values stored by TLS.
	================================================================================	*/
void doTaskJob (int* datas)
	{
	int i, val ;
	char* s ;
	bigtime_t smallDelay = rand()>>3 ;
	
	for (i=0 ; i<NUMBER_OF_INDEXS ; i++)
		{
		snooze (smallDelay) ;
		assert (B_NO_ERROR == tls_get (indexs[i], &val)) ;
		assert (val == datas[i]) ;
		}
	}


/*	===== entryProc ================================================================
	The entry proc.
	Normally called before first use of TLS data access services in a thread.
	================================================================================	*/
void entryProc (thread_id id, int param, void* threadEnv)
	{
	char* s ;
	
	printf ("entryProc called in thread %d for thread %d.\n", find_thread(NULL), id) ;
	s = (char*) malloc (20) ;
	sprintf (s, "Thread(%d)", (int)id) ;
	assert (B_NO_ERROR == tls_set_inproc (param, (tls_data_t)s, threadEnv)) ;
	}
	
	
/*	===== entryProc ================================================================
	The exit proc.
	Normally called after last use of TLS data access services in a thread.
	================================================================================	*/
void exitProc (thread_id id, int param, void* threadEnv)
	{
	void *p ;

	printf ("exitProc called in thread %d for thread %d.\n", find_thread(NULL), id) ;
	assert (B_NO_ERROR == tls_get_inproc (param, (tls_data_t*)&p, threadEnv)) ;
	free (p) ;
	}

